const ttf2svg = require('ttf2svg')
const css2json = require('css2json')
const request = require('request')
const fsExtra = require('fs-extra')
const fs = require('fs')
const path = require('path')
const process = require('process')

const Base64={_keyStr:"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=",encode:function(e){var t="";var n,r,i,s,o,u,a;var f=0;e=Base64._utf8_encode(e);while(f<e.length){n=e.charCodeAt(f++);r=e.charCodeAt(f++);i=e.charCodeAt(f++);s=n>>2;o=(n&3)<<4|r>>4;u=(r&15)<<2|i>>6;a=i&63;if(isNaN(r)){u=a=64;}else if(isNaN(i)){a=64;}t=t+this._keyStr.charAt(s)+this._keyStr.charAt(o)+this._keyStr.charAt(u)+this._keyStr.charAt(a);}return t;},decode:function(e){var t="";var n,r,i;var s,o,u,a;var f=0;e=e.replace(/[^A-Za-z0-9\+\/\=]/g,"");while(f<e.length){s=this._keyStr.indexOf(e.charAt(f++));o=this._keyStr.indexOf(e.charAt(f++));u=this._keyStr.indexOf(e.charAt(f++));a=this._keyStr.indexOf(e.charAt(f++));n=s<<2|o>>4;r=(o&15)<<4|u>>2;i=(u&3)<<6|a;t=t+String.fromCharCode(n);if(u!=64){t=t+String.fromCharCode(r);}if(a!=64){t=t+String.fromCharCode(i);}}t=Base64._utf8_decode(t);return t;},_utf8_encode:function(e){e=e.replace(/\r\n/g,"\n");var t="";for(var n=0;n<e.length;n++){var r=e.charCodeAt(n);if(r<128){t+=String.fromCharCode(r);}else if(r>127&&r<2048){t+=String.fromCharCode(r>>6|192);t+=String.fromCharCode(r&63|128);}else{t+=String.fromCharCode(r>>12|224);t+=String.fromCharCode(r>>6&63|128);t+=String.fromCharCode(r&63|128);}}return t;},_utf8_decode:function(e){var t="";var n=0;var r=c1=c2=0;while(n<e.length){r=e.charCodeAt(n);if(r<128){t+=String.fromCharCode(r);n++;}else if(r>191&&r<224){c2=e.charCodeAt(n+1);t+=String.fromCharCode((r&31)<<6|c2&63);n+=2;}else{c2=e.charCodeAt(n+1);c3=e.charCodeAt(n+2);t+=String.fromCharCode((r&15)<<12|(c2&63)<<6|c3&63);n+=3;}}return t;}};

function svgToDataUrl(str) {
  return `data:image/svg+xml;base64,${Base64.encode(str)}`;
}

function pathToSvg(path, fill, width) {
  return `<svg viewBox='0 -128 1035 1035' width='${width}' xmlns='http://www.w3.org/2000/svg' style='transform:rotateX(180deg);transform-origin:center;scale(.2);'><path fill='${fill}' d='${path}'></path></svg>`;
}

function parseIconfont(svgContent, fill, width, svgType) {
  // 构建单个 icon 的数据结构
  const singleIconRegex = /<glyph glyph-name=\"(.*?)\" unicode=\"(.*?)\" d=\"(.*?)\".*\/>/g;
  let codeMap  = {};
  let match = singleIconRegex.exec(svgContent);
  while (match) {
    const [, ,code, path] = match;
    // 解析到的 code 是 10 进制，代码里用的 code 是 16 进制，为了方便匹配做个转换
    const finalName = code.toLowerCase().replace(/[&|#|x|;]/g, '');
    codeMap[finalName] = svgType === 'base64' ? svgToDataUrl(pathToSvg(path, fill, width)) : pathToSvg(path, fill, width);
    match = singleIconRegex.exec(svgContent);
  }
  return codeMap;
}

function generateStep(fontKey, ttfUrl, cssUrl, fill, width, svgType) {
  return new Promise((resolve, reject) => {
      // 第一步：ttf -> svg
      request(ttfUrl)
      .pipe(fsExtra.createWriteStream(`.iconfont/${fontKey}.ttf`))
      .on('close', function () {
        console.log(`${fontKey} ttf written!`);
        fs.readFile(`.iconfont/${fontKey}.ttf`, function (err, buffer) {
          try {
            // svg 数据
            const svgContent = ttf2svg(buffer);
            // icon 结构化数据
            const unicodeMaps = parseIconfont(svgContent, fill, width, svgType);
              // 第二步：css -> json
              request(cssUrl)
              .pipe(fs.createWriteStream(`.iconfont/${fontKey}.css`))
              .on('close', function () {
                console.log(`${fontKey} css written!`);
                fs.readFile(`.iconfont/${fontKey}.css`, 'utf-8', function (err, buffer) {
                  try {
                    // css 转成 json
                    const cssJson = css2json(buffer.toString());
                    // 最终输出的数据定义
                    const schemaData = {};
                    // unicode 和 css content 匹配逻辑
                    for(const classKey in cssJson) {
                      if(classKey.endsWith(":before") && Object.keys(cssJson[classKey]).length === 1 && 'content' in cssJson[classKey]) {
                        const iconUnicode = cssJson[classKey].content.replace(/\"\\|\"/g, "");
                        if(unicodeMaps[iconUnicode]) {
                          const iconClass = classKey.replace(/\:before|\./ig, "");
                          schemaData[iconClass] = cssJson[classKey];
                          schemaData[iconClass]['svg'] = unicodeMaps[iconUnicode];
                          schemaData[iconClass]['type'] = fontKey;
                        }
                      }
                    }
                    // 删除文件
                    fs.unlinkSync(`.iconfont/${fontKey}.ttf`)
                    fs.unlinkSync(`.iconfont/${fontKey}.css`)
                    resolve(schemaData)
                  } catch(ex) {
                    fs.unlinkSync(`.iconfont/${fontKey}.css`)
                    reject('css地址有问题，请检查！');
                  }
                });
              })
          } catch(ex) {
            fs.unlinkSync(`.iconfont/${fontKey}.ttf`)
            reject('ttf地址有问题，请检查！');
          }
        });
      });
  })
} 

function iconfontExport({ configPath, outputDir = ".iconfont", fill = "#ffffff", width = "10px", svgType = "svg" }) {
  return new Promise((resolve, reject) => {
    fsExtra.ensureDir(outputDir);
    fs.readFile(path.resolve(process.cwd(), configPath), 'utf-8', function (err, data) {
      const iconfontConfigs = JSON.parse(data);
      const promiseArr = [];
      for(const iconKey in iconfontConfigs) {
        const { ttf, css } = iconfontConfigs[iconKey];
        if(ttf && css) {
          promiseArr.push(generateStep(iconKey, ttf, css, fill, width, svgType));
        }
      };
      Promise.all(promiseArr).then(schemaList => {
        const data = schemaList.reduce((prev, next) => ({...prev, ...next }), {});
        fsExtra.writeFile(`${outputDir}/schema.json`, JSON.stringify(data, null, 4));
        resolve(data);
        console.log("生成完毕!")
      }).catch(e=> {
        reject('生成失败:', e);
        console.log("生成失败:", e);
      });
    })
  })
}

exports.svgToDataUrl = svgToDataUrl;

exports.iconfontExport = iconfontExport;